import { Chessboard, isEmpty, Square, squareAtPosition } from "./chessboard";
import { Move } from "./movements";
import { equals, left, right, top, bottom, Position} from "./position";

/**
 * Checks whether a Black Pawn can perform a given move.
 * A pawn can move forward to the unoccupied square immediately in front of 
 * it on the same file, or on its first move it can advance two squares along 
 * the same file, provided both squares are unoccupied (black dots in the 
 * diagram); or the pawn can capture an opponent's piece on a square diagonally 
 * in front of it on an adjacent file, by moving to that square (black "x"s). 
 * 
 * 
 * @param board The chessboard of the current game
 * @param move 
 */
export function blackPawnMove(board: Chessboard, move: Move): boolean {

    if (equals(move.to!, bottom(move.from!))) {
        //console.log("Single forward");
        return isEmpty(board, move.to!);
    }

    if (move.from!.rank === 6 && equals(move.to!, bottom(bottom(move.from!)))) {
        //console.log("Double forward");
        return isEmpty(board, bottom(move.from!)) && isEmpty(board, move.to!);
    }

    if (equals(move.to!, left(bottom(move.from!))) || equals(move.to!, right(bottom(move.from!)))) {
        let destination: Square = squareAtPosition(board, move.to!);
        return !(destination.isEmpty || !destination.piece!.isWhite)
    }

    return false;
}

/**
 * A pawn can move forward to the unoccupied square immediately in front of 
 * it on the same file, or on its first move it can advance two squares along 
 * the same file, provided both squares are unoccupied (black dots in 
 * the diagram); or the pawn can capture an opponent's piece on a square diagonally 
 * in front of it on an adjacent file, by moving to that square (black "x"s). 
 * 
 * 
 * @param board The chessboard of the current game
 * @param move 
 */
export function whitePawnMove(board: Chessboard, move: Move): boolean {

    if (equals(move.to!, top(move.from!))) {
        return isEmpty(board, move.to!);
    }

    if (move.from!.rank === 1 && equals(move.to!, top(top(move.from!)))) {
        return isEmpty(board, top(move.from!)) && isEmpty(board, move.to!);
    }

    if (equals(move.to!, left(top(move.from!))) || equals(move.to!, right(top(move.from!)))) {
        let destination: Square = squareAtPosition(board, move.to!);
        return !(destination.isEmpty || destination.piece!.isWhite)
    }

    return false;
}

/**
 * Checks whether a King can perform a given move.
 * The king moves one square in any direction. 
 * 
 * @param board The chessboard of the current game
 * @param move 
 */
export function kingMove(board: Chessboard, move: Move): boolean {
    let destination: Square = squareAtPosition(board, move.to!);
    let source: Square = squareAtPosition(board, move.from!);

    if (equals(move.to!, top(move.from!))) {

        if (isEmpty(board, move.to!)) {
            return true;
        } else {
            return !(isEmpty(board, move.to!) || source.piece!.isWhite == destination.piece!.isWhite);
        }
    }
    if (equals(move.to!, bottom(move.from!))) {

        if (isEmpty(board, move.to!)) {
            return true;
        } else {
            return !(isEmpty(board, move.to!) || source.piece!.isWhite == destination.piece!.isWhite);
        }
    }
    if (equals(move.to!, right(move.from!))) {

        if (isEmpty(board, move.to!)) {
            return true;
        } else {
            return !(isEmpty(board, move.to!) || source.piece!.isWhite == destination.piece!.isWhite);
        }
    }
    if (equals(move.to!, left(move.from!))) {

        if (isEmpty(board, move.to!)) {
            return true;
        } else {
            return !(isEmpty(board, move.to!) || source.piece!.isWhite == destination.piece!.isWhite);
        }
    }
    if (equals(move.to!, top(left(move.from!)))) {

        if (isEmpty(board, move.to!)) {
            return true;
        } else {
            return !(isEmpty(board, move.to!) || source.piece!.isWhite == destination.piece!.isWhite);
        }
    }
    if (equals(move.to!, bottom(left(move.from!)))) {

        if (isEmpty(board, move.to!)) {
            return true;
        } else {
            return !(isEmpty(board, move.to!) || source.piece!.isWhite == destination.piece!.isWhite);
        }
    }
    if (equals(move.to!, top(right(move.from!)))) {

        if (isEmpty(board, move.to!)) {
            return true;
        } else {
            return !(isEmpty(board, move.to!) || source.piece!.isWhite == destination.piece!.isWhite);
        }
    }
    if (equals(move.to!, bottom(right(move.from!)))) {

        if (isEmpty(board, move.to!)) {
            return true;
        } else {
            return !(isEmpty(board, move.to!) || source.piece!.isWhite == destination.piece!.isWhite);
        }
    }

    return false;
}

/**
 * Checks whether a Queen can perform a given move.
 * The queen combines the power of a rook and bishop and can move any 
 * number of squares along a rank, file, or diagonal, but cannot leap over other pieces.
 * 
 * @param board The chessboard of the current game
 * @param move 
 */
export function queenMove(board: Chessboard, move: Move): boolean {
    let destination: Square = squareAtPosition(board, move.to!);
    let source: Square = squareAtPosition(board, move.from!);

    //check that the movement goes from right to left
    if ((move.from!.rank == move.to!.rank) && (move.to!.file < move.from!.file)) {
        let newpos: Position = { rank: move.from!.rank, file: move.from!.file - 1 };
        
        while ((newpos.file > 0) && !equals(newpos, move.to!) && isEmpty(board, newpos)) {
            newpos = left(newpos);
        }
        if (equals(newpos, move.to!) && isEmpty(board, newpos)) {
            return true;
        } else if (equals(newpos, move.to!) && !isEmpty(board, newpos)) {
            if (source.piece!.isWhite == destination.piece!.isWhite) {
                return false;
            } else {
                return true;
            }
        } else {
            return false;
        }
    //check that the movement goes from left to right
    } else if ((move.from!.rank == move.to!.rank) && (move.to!.file > move.from!.file)) {
        let newpos: Position = { rank: move.from!.rank, file: move.from!.file + 1 };

        while ((newpos.file < 7) && !equals(newpos, move.to!) && isEmpty(board, newpos)) {
            newpos = right(newpos);
        }
        if (equals(newpos, move.to!) && isEmpty(board, newpos)) {
            return true;
        } else if (equals(newpos, move.to!) && !isEmpty(board, newpos)) {

            if (source.piece!.isWhite == destination.piece!.isWhite) {
                return false;
            } else {
                return true;
            }
        } else {
            return false;
        }
    //check that the movement goes from bottom to top
    } else if ((move.from!.file == move.to!.file) && (move.to!.rank > move.from!.rank)) {
        let newpos: Position = { rank: move.from!.rank + 1, file: move.from!.file };

        while ((newpos.rank < 7) && !equals(newpos, move.to!) && isEmpty(board, newpos)) {
            newpos = top(newpos);
        }
        if (equals(newpos, move.to!) && isEmpty(board, newpos)) {
            return true;
        } else if (equals(newpos, move.to!) && !isEmpty(board, newpos)) {

            if (source.piece!.isWhite == destination.piece!.isWhite) {
                return false;
            } else {
                return true;
            }
        } else {
            return false;
        }
        //check that the movement goes from top to bottom
    } else if ((move.from!.file == move.to!.file) && (move.to!.rank < move.from!.rank)) {
        let newpos: Position = { rank: move.from!.rank - 1, file: move.from!.file };

        while ((newpos.rank > 0) && !equals(newpos, move.to!) && isEmpty(board, newpos)) {
            newpos = bottom(newpos);
        }
        if (equals(newpos, move.to!) && isEmpty(board, newpos)) {
            return true;
        } else if (equals(newpos, move.to!) && !isEmpty(board, newpos)) {

            if (source.piece!.isWhite == destination.piece!.isWhite) {
                return false;
            } else {
                return true;
            }
        } else {
            return false;
        }
        //check that the movement goes from top right to bottom left
    } else if ((move.to!.file < move.from!.file) && (move.to!.rank < move.from!.rank)) {
        let newpos: Position = { rank: move.from!.rank - 1, file: move.from!.file - 1 };

        while ((newpos.rank > 0) && (newpos.file > 0) && !equals(newpos, move.to!) && isEmpty(board, newpos)) {
            newpos = bottom(left(newpos));
        }
        if (equals(newpos, move.to!) && isEmpty(board, newpos)) {
            return true;
        } else if (equals(newpos, move.to!) && !isEmpty(board, newpos)) {

            if (source.piece!.isWhite == destination.piece!.isWhite) {
                return false;
            } else {
                return true;
            }
        } else {
            return false;
        }
    //check that the movement goes from the bottom top left to bottom right
    } else if ((move.to!.file > move.from!.file) && (move.to!.rank < move.from!.rank)) {
        let newpos: Position = { rank: move.from!.rank - 1, file: move.from!.file + 1 };

        while ((newpos.rank > 0) && (newpos.file < 7) && !equals(newpos, move.to!) && isEmpty(board, newpos)) {
            newpos = bottom(right(newpos));
        }
        if (equals(newpos, move.to!) && isEmpty(board, newpos)) {
            return true;
        } else if (equals(newpos, move.to!) && !isEmpty(board, newpos)) {

            if (source.piece!.isWhite == destination.piece!.isWhite) {
                return false;
            } else {
                return true;
            }
        } else {
            return false;
        }
    //check that the movement goes from the bottom left to top right
    } else if ((move.to!.file > move.from!.file) && (move.to!.rank > move.from!.rank)) {
        let newpos: Position = { rank: move.from!.rank + 1, file: move.from!.file + 1 };

        while ((newpos.rank < 7) && (newpos.file < 7) && !equals(newpos, move.to!) && isEmpty(board, newpos)) {
            newpos = top(right(newpos));
        }
        if (equals(newpos, move.to!) && isEmpty(board, newpos)) {
            return true;
        } else if (equals(newpos, move.to!) && !isEmpty(board, newpos)) {

            if (source.piece!.isWhite == destination.piece!.isWhite) {
                return false;
            } else {
                return true;
            }
        } else {
            return false;
        }
    //check that the movement goes from the bottom right to top left
    } else if ((move.to!.file < move.from!.file) && (move.to!.rank > move.from!.rank)) {
        let newpos: Position = { rank: move.from!.rank + 1, file: move.from!.file - 1 };

        while ((newpos.rank < 7) && (newpos.file > 0) && !equals(newpos, move.to!) && isEmpty(board, newpos)) {
            newpos = top(left(newpos));
        }
        if (equals(newpos, move.to!) && isEmpty(board, newpos)) {
            return true;
        } else if (equals(newpos, move.to!) && !isEmpty(board, newpos)) {

            if (source.piece!.isWhite == destination.piece!.isWhite) {
                return false;
            } else {
                return true;
            }
        } else {
            return false;
        }
    }
    return false;
}

/**
 * Checks whether a Empress can perform a given move.
 * An Empress can move any number of squares along a rank or file, 
 * but cannot leap over other pieces.
 * An Empress can also move to any of the closest squares that are not on the 
 * same rank, file, or diagonal. (Thus the move forms an "L"-shape: 
 * two squares vertically and one square horizontally, or two 
 * squares horizontally and one square vertically.) 
 * 
 * @param board The chessboard of the current game
 * @param move 
 */
export function empressMove(board: Chessboard, move: Move, isWhite: Boolean): boolean {

    let numberOfRank: number = move.to!.rank - move.from!.rank
    let numberOfFile: number = move.to!.file - move.from!.file

    // Check which movements is playing
    if (numberOfFile != 0 && numberOfRank != 0) {
       
        let absNumberOfFile: number = Math.abs(numberOfFile)
        let absNumberOfRank: number = Math.abs(numberOfRank)

        // Check for poney movement
        if (absNumberOfRank == 2 && absNumberOfFile == 1) {

            if (numberOfRank == 2) {
                for (let i: number = 1; i <= 2; i++) {
                    let square: Square = squareAtPosition(board, {file: move.from!.file, rank: move.from!.rank + (1 * i)})
    
                    if (!square.isEmpty) {
                        return false
                    }
                }
            } else if (numberOfRank == -2) {
                for (let i: number = 1; i <= 2; i++) {
                    let square: Square = squareAtPosition(board, {file: move.from!.file, rank: move.from!.rank - (1 * i)})
    
                    if (!square.isEmpty) {
                        return false
                    }
                }
            }

            let destination: Square = squareAtPosition(board, move.to!)

            if (destination.isEmpty) {
                return true
            } else {
                return isWhite == destination.piece?.isWhite ? false : true
            }

        } else if (absNumberOfRank == 1 && absNumberOfFile == 2) {

            if (numberOfFile == 2) {
                for (let i: number = 1; i <= 2; i++) {
                    let square: Square = squareAtPosition(board, {file: move.from!.file + (1 * i), rank: move.from!.rank})
    
                    if (!square.isEmpty) {
                        return false
                    }
                }
            } else if (numberOfFile == -2) {
                for (let i: number = 1; i <= 2; i++) {
                    let square: Square = squareAtPosition(board, {file: move.from!.file - (1 * i), rank: move.from!.rank})
    
                    if (!square.isEmpty) {
                        return false
                    }
                }
            }

            let destination: Square = squareAtPosition(board, move.to!)

            if (destination.isEmpty) {
                return true
            } else {
                return isWhite == destination.piece?.isWhite ? false : true
            }
        }

    } else {
        // Tower movement

        if (numberOfRank < 0) {
            //  Down deplacement

            for (let i:number = -1; i >= numberOfRank; i--) {
        
                let square: Square = squareAtPosition(board, {file: move.from!.file, rank: move.from!.rank + (1 * i)})

                if (equals(move.to!, square.position)) {
                    // The final case
                    if (square.isEmpty) {
                        return true
                    } else {
                        return isWhite == square.piece?.isWhite ? false : true
                    }

                } else if (!square.isEmpty) {
                    // There is another pawn on is trajectory
                    return false
                }
            }

        } else if (numberOfRank > 0) {
            // Up deplacement
 
            for (let i:number = 1; i <= numberOfRank; i++) {

                let square: Square = squareAtPosition(board, {file: move.from!.file, rank: move.from!.rank + (1 * i)})

                if (equals(move.to!, square.position)) {
                    // The final case
                    if (square.isEmpty) {
                        return true
                    } else {
                        return isWhite == square.piece?.isWhite ? false : true
                    }

                } else if (!square.isEmpty) {
                    // There is another pawn on is trajectory
                    return false
                }
            }

        } else if (numberOfFile < 0) {
            // Left deplacement
            
            for (let i:number = -1; i >= numberOfFile; i--) {
                let square: Square = squareAtPosition(board, {file: move.from!.file + (1 * i), rank: move.from!.rank})

                if (equals(move.to!, square.position)) {
                    // The final case
                    if (square.isEmpty) {
                        return true
                    } else {
                        return isWhite == square.piece?.isWhite ? false : true
                    }

                } else if (!square.isEmpty) {
                    // There is another pawn on is trajectory
                    return false
                }
            }

        } else if (numberOfFile > 0) {
            // Right deplacement
  
            for (let i:number = 1; i <= numberOfFile; i++) {
                let square: Square = squareAtPosition(board, {file: move.from!.file + (1 * i), rank: move.from!.rank})

                if (equals(move.to!, square.position)) {
                    // The final case
                    if (square.isEmpty) {
                        return true
                    } else {
                        return isWhite == square.piece?.isWhite ? false : true
                    }

                } else if (!square.isEmpty) {
                    // There is another pawn on is trajectory
                    return false
                }
            }
        }

    }

    return false;
}

/**
 * Checks whether a Princess can perform a given move.
 * A princess can move any number of squares diagonally, 
 * but cannot leap over other pieces.
 * A princess can also move to any of the closest squares that are not on the 
 * same rank, file, or diagonal. (Thus the move forms an "L"-shape: 
 * two squares vertically and one square horizontally, or two 
 * squares horizontally and one square vertically.) 
 * 
 * @param board The chessboard of the current game
 * @param move 
 */
export function princessMove(board: Chessboard, move: Move): boolean {
    let destination: Square = squareAtPosition(board, move.to!);
    let source: Square = squareAtPosition(board, move.from!);

    if (equals(move.to!, right(top(top(move.from!))))) {

        if (isEmpty(board, move.to!)) {
            return true;
        } else {
            return !(isEmpty(board, move.to!) || source.piece!.isWhite == destination.piece!.isWhite);
        }
    }
    if (equals(move.to!, left(top(top(move.from!))))) {

        if (isEmpty(board, move.to!)) {
            return true;
        } else {
            return !(isEmpty(board, move.to!) || source.piece!.isWhite == destination.piece!.isWhite);
        }
    }
    if (equals(move.to!, right(right(top(move.from!))))) {

        if (isEmpty(board, move.to!)) {
            return true;
        } else {
            return !(isEmpty(board, move.to!) || source.piece!.isWhite == destination.piece!.isWhite);
        }
    }
    if (equals(move.to!, right(right(bottom(move.from!))))) {

        if (isEmpty(board, move.to!)) {
            return true;
        } else {
            return !(isEmpty(board, move.to!) || source.piece!.isWhite == destination.piece!.isWhite);
        }
    }
    if (equals(move.to!, bottom(bottom(right(move.from!))))) {

        if (isEmpty(board, move.to!)) {
            return true;
        } else {
            return !(isEmpty(board, move.to!) || source.piece!.isWhite == destination.piece!.isWhite);
        }
    }
    if (equals(move.to!, bottom(bottom(left(move.from!))))) {

        if (isEmpty(board, move.to!)) {
            return true;
        } else {
            return !(isEmpty(board, move.to!) || source.piece!.isWhite == destination.piece!.isWhite);
        }
    }
    if (equals(move.to!, top(left(left(move.from!))))) {

        if (isEmpty(board, move.to!)) {
            return true;
        } else {
            return !(isEmpty(board, move.to!) || source.piece!.isWhite == destination.piece!.isWhite);
        }
    }
    if (equals(move.to!, bottom(left(left(move.from!))))) {

        if (isEmpty(board, move.to!)) {
            return true;
        } else {
            return !(isEmpty(board, move.to!) || source.piece!.isWhite == destination.piece!.isWhite);
        }
    }
    //check that the movement goes from top right to bottom left
    if ((move.to!.file < move.from!.file) && (move.to!.rank < move.from!.rank)) {
        let newpos: Position = { rank: move.from!.rank - 1, file: move.from!.file - 1 };

        while ((newpos.rank > 0) && (newpos.file > 0) && !equals(newpos, move.to!) && isEmpty(board, newpos)) {
            newpos = bottom(left(newpos));
        }
        if (equals(newpos, move.to!) && isEmpty(board, newpos)) {
            return true;
        } else if (equals(newpos, move.to!) && !isEmpty(board, newpos)) {

            if (source.piece!.isWhite == destination.piece!.isWhite) {
                return false;
            } else {
                return true;
            }
        } else {
            return false;
        }
    //check that the movement goes from top left to bottom right
    } else if ((move.to!.file > move.from!.file) && (move.to!.rank < move.from!.rank)) {
        let newpos: Position = { rank: move.from!.rank - 1, file: move.from!.file + 1 };

        while ((newpos.rank > 0) && (newpos.file < 7) && !equals(newpos, move.to!) && isEmpty(board, newpos)) {
            newpos = bottom(right(newpos));
        }
        if (equals(newpos, move.to!) && isEmpty(board, newpos)) {
            return true;
        } else if (equals(newpos, move.to!) && !isEmpty(board, newpos)) {

            if (source.piece!.isWhite == destination.piece!.isWhite) {
                return false;
            } else {
                return true;
            }
        } else {
            return false;
        }
    //check that the movement goes from bottom left to top right
    } else if ((move.to!.file > move.from!.file) && (move.to!.rank > move.from!.rank)) {
        let newpos: Position = { rank: move.from!.rank + 1, file: move.from!.file + 1 };

        while ((newpos.rank < 7) && (newpos.file < 7) && !equals(newpos, move.to!) && isEmpty(board, newpos)) {
            newpos = top(right(newpos));
        }
        if (equals(newpos, move.to!) && isEmpty(board, newpos)) {
            return true;
        } else if (equals(newpos, move.to!) && !isEmpty(board, newpos)) {

            if (source.piece!.isWhite == destination.piece!.isWhite) {
                return false;
            } else {
                return true;
            }
        } else {
            return false;
        }
    //check that the movement goes from bottom right to top left
    } else if ((move.to!.file < move.from!.file) && (move.to!.rank > move.from!.rank)) {
        let newpos: Position = { rank: move.from!.rank + 1, file: move.from!.file - 1 };

        while ((newpos.rank < 7) && (newpos.file > 0) && !equals(newpos, move.to!) && isEmpty(board, newpos)) {
            newpos = top(left(newpos));
        }
        if (equals(newpos, move.to!) && isEmpty(board, newpos)) {
            return true;
        } else if (equals(newpos, move.to!) && !isEmpty(board, newpos)) {

            if (source.piece!.isWhite == destination.piece!.isWhite) {
                return false;
            } else {
                return true;
            }
        } else {
            return false;
        }
    }
    return false;
}

/**
 * Checks whether a Camel can perform a given move.
 * The Camel move forms an "L"-shape: 
 * three squares vertically and one square horizontally, or three 
 * squares horizontally and one square vertically.) 
 * 
 * The camel can leap over other pieces.
 * 
 * @param board The chessboard of the current game
 * @param move 
 */
export function camelMove(board: Chessboard, move: Move, isWhite: Boolean): boolean {

    if (equals(move.to!, top(top(top(right(move.from!)))))) {
        let destination: Square = squareAtPosition(board, move.to!)

        if (isWhite) {
            return (destination.isEmpty || !destination.piece!.isWhite)
        } else {
            return (destination.isEmpty || destination.piece!.isWhite)
        }
    } 

    if (equals(move.to!, top(top(top(left(move.from!)))))) {
        let destination: Square = squareAtPosition(board, move.to!)

        if (isWhite) {
            return (destination.isEmpty || !destination.piece!.isWhite)
        } else {
            return (destination.isEmpty || destination.piece!.isWhite)
        }
    }

    if (equals(move.to!, bottom(bottom(bottom(left(move.from!)))))) {
        let destination: Square = squareAtPosition(board, move.to!)

        if (isWhite) {
            return (destination.isEmpty || !destination.piece!.isWhite)
        } else {
            return (destination.isEmpty || destination.piece!.isWhite)
        }
    }

    if (equals(move.to!, bottom(bottom(bottom(right(move.from!)))))) {
        let destination: Square = squareAtPosition(board, move.to!)

        if (isWhite) {
            return (destination.isEmpty || !destination.piece!.isWhite)
        } else {
            return (destination.isEmpty || destination.piece!.isWhite)
        }
    }

    if (equals(move.to!, left(left(left(bottom(move.from!)))))) {
        let destination: Square = squareAtPosition(board, move.to!)

        if (isWhite) {
            return (destination.isEmpty || !destination.piece!.isWhite)
        } else {
            return (destination.isEmpty || destination.piece!.isWhite)
        }
    }

    if (equals(move.to!, left(left(left(top(move.from!)))))) {
        let destination: Square = squareAtPosition(board, move.to!)

        if (isWhite) {
            return (destination.isEmpty || !destination.piece!.isWhite)
        } else {
            return (destination.isEmpty || destination.piece!.isWhite)
        }
    }

    if (equals(move.to!, right(right(right(bottom(move.from!)))))) {
        let destination: Square = squareAtPosition(board, move.to!)

        if (isWhite) {
            return (destination.isEmpty || !destination.piece!.isWhite)
        } else {
            return (destination.isEmpty || destination.piece!.isWhite)
        }
    }

    if (equals(move.to!, right(right(right(top(move.from!)))))) {
        let destination: Square = squareAtPosition(board, move.to!)

        if (isWhite) {
            return (destination.isEmpty || !destination.piece!.isWhite)
        } else {
            return (destination.isEmpty || destination.piece!.isWhite)
        }
    }

    return false;
} 
