import { Expect, Test, Setup} from "alsatian";
import { Chessboard, createEmptyChessboard, putPiece } from '../../main/ts/chessboard';
import * as isPossible from '../../main/ts/move-validation'
import * as pieces from '../../main/ts/piece'
import { Position, position } from '../../main/ts/position';
import { Move, move } from '../../main/ts/movements';

const positionA8 : Position = position(0, 7) // A8

const positionB3 : Position = position(1, 2) // B3
const positionB5 : Position = position(1, 4) // B5

const positionC4 : Position = position(2, 3) // C4

const positionD1 : Position = position(3, 0) // D1
const positionD7 : Position = position(3, 6) // D7

const positionE1 : Position = position(4, 0) // E1
const positionE4 : Position = position(4, 3) // E4

const positionF1 : Position = position(5, 0) // F1
const positionF7 : Position = position(5, 6) // F7
const positionH3 : Position = position(7, 2) // H3
const positionH4 : Position = position(7, 3) // H4

// Horizontal moves
const moveE4_H4 : Move = move(positionE4, positionH4);

// Vertical moves
const moveE4_E1 : Move = move(positionE4, positionE1);

// Diagonal moves
const moveE4_A8 : Move = move(positionE4, positionA8);

// Camel moves
const moveE4_F7 : Move = move(positionE4, positionF7);
const moveE4_F1 : Move = move(positionE4, positionF1);
const moveE4_H3 : Move = move(positionE4, positionH3);
const moveE4_D1 : Move = move(positionE4, positionD1);
const moveE4_B3 : Move = move(positionE4, positionB3);
const moveE4_B5 : Move = move(positionE4, positionB5);
const moveE4_D7 : Move = move(positionE4, positionD7);

let chessboard : Chessboard;

export class TestCamelMoves {
    @Setup
    beforeEach() {
        // Initialize an empty chessboard
        chessboard = createEmptyChessboard()

        // Place a white Camel on E4
        putPiece(chessboard, positionE4, pieces.whiteCamel)
    }

    @Test("A Camel can move three squares horizontally and one square vertically")
    testCanMoveThreeHorizontalAndOneVertical() {
        // Check the following moves are possible:
        // - moveE4_H3
        Expect(isPossible.camelMove(chessboard, moveE4_H3, true)).toBeTruthy()

        // - moveE4_B3
        Expect(isPossible.camelMove(chessboard, moveE4_B3, true)).toBeTruthy()

        // - moveE4_B5
        Expect(isPossible.camelMove(chessboard, moveE4_B5, true)).toBeTruthy()
    }

    @Test("A Camel can move three squares vertically and one square horizontally")
    testCanMoveThreeVerticalAndOneHorizontal() {
        // Check the following moves are possible:
        // - moveE4_F1
        Expect(isPossible.camelMove(chessboard, moveE4_F1, true)).toBeTruthy()

        // - moveE4_F7
        Expect(isPossible.camelMove(chessboard, moveE4_F7, true)).toBeTruthy()

        // - moveE4_D1
        Expect(isPossible.camelMove(chessboard, moveE4_D1, true)).toBeTruthy()

        // - moveE4_D7
        Expect(isPossible.camelMove(chessboard, moveE4_D7, true)).toBeTruthy()
    }

    @Test("A Camel can leap other pieces")
    testCanLeapOtherPieces() {
        // Place a black Pawn on C4
        putPiece(chessboard, positionC4, pieces.blackPawn)

        // Check the move moveE4_B3 is possible
        Expect(isPossible.camelMove(chessboard, moveE4_B3, true)).toBeTruthy()
    }

    @Test("A Camel cannot move diagonally")
    testCannotMoveDiagonally() {
        // Check the move moveE4_A8 is impossible
        Expect(isPossible.camelMove(chessboard, moveE4_A8, true)).not.toBeTruthy()
    }

    @Test("A Camel cannot move horizontally")
    testCannotMoveHorizontally() {
        // Check the move moveE4_H4 is impossible
        Expect(isPossible.camelMove(chessboard, moveE4_H4, true)).not.toBeTruthy()
    }

    @Test("A Camel cannot move vertically")
    testCannotMoveVertically() {
        // Check the move moveE4_E1 is impossible
        Expect(isPossible.camelMove(chessboard, moveE4_E1, true)).not.toBeTruthy()
    }

    @Test("A Camel can capture a piece from another color")
    testCanCaptureAnotherColor() {
        // Place a black Pawn on H3
        putPiece(chessboard, positionH3, pieces.blackPawn)

        // Check the move moveE4_H3 is possible
        Expect(isPossible.camelMove(chessboard, moveE4_H3, true)).toBeTruthy()
    }

    @Test("A Camel cannot capture a piece from the same color")
    testCannotCaptureSameColor() {
        // Place a white Pawn on H3
        putPiece(chessboard, positionH3, pieces.whitePawn)

        // Check the move moveE4_H3 is possible
        Expect(isPossible.camelMove(chessboard, moveE4_H3, true)).not.toBeTruthy()
    }
}
