import { Expect, Test, Setup} from "alsatian";
import { Chessboard, createEmptyChessboard, putPiece } from '../../main/ts/chessboard';
import * as isPossible from '../../main/ts/move-validation'
import * as pieces from '../../main/ts/piece'
import { Position, position } from '../../main/ts/position';
import { Move, move } from '../../main/ts/movements';

const positionA4 : Position = position(0, 3) // A4
const positionA8 : Position = position(0, 7) // A8

const positionB1 : Position = position(1, 0) // B1

const positionC3 : Position = position(2, 2) // C3
const positionC5 : Position = position(2, 4) // C5

const positionD2 : Position = position(3, 1) // D2
const positionD6 : Position = position(3, 5) // D6

const positionE1 : Position = position(4, 0) // E1
const positionE3 : Position = position(4, 2) // E3
const positionE4 : Position = position(4, 3) // E4
const positionE8 : Position = position(4, 7) // E8

const positionF2 : Position = position(5, 1) // F2
const positionF4 : Position = position(5, 3) // F4
const positionF6 : Position = position(5, 5) // F6

const positionG3 : Position = position(6, 2) // G3
const positionG5 : Position = position(6, 4) // G5

const positionH1 : Position = position(7, 0) // H1
const positionH4 : Position = position(7, 3) // H4
const positionH7 : Position = position(7, 6) // H7

// Horizontal moves
const moveE4_H4 : Move = move(positionE4, positionH4);
const moveE4_A4 : Move = move(positionE4, positionA4);

// Vertical moves
const moveE4_E1 : Move = move(positionE4, positionE1);
const moveE4_E8 : Move = move(positionE4, positionE8);

// Diagonal moves
const moveE4_A8 : Move = move(positionE4, positionA8);
const moveE4_B1 : Move = move(positionE4, positionB1);
const moveE4_H7 : Move = move(positionE4, positionH7);
const moveE4_H1 : Move = move(positionE4, positionH1);

// Empress moves
const moveE4_G3 : Move = move(positionE4, positionG3);
const moveE4_G5 : Move = move(positionE4, positionG5);
const moveE4_C3 : Move = move(positionE4, positionC3);
const moveE4_C5 : Move = move(positionE4, positionC5);
const moveE4_F2 : Move = move(positionE4, positionF2);
const moveE4_F6 : Move = move(positionE4, positionF6);
const moveE4_D2 : Move = move(positionE4, positionD2);
const moveE4_D6 : Move = move(positionE4, positionD6);

let chessboard : Chessboard;

export class TestEmpressMoves {
    @Setup
    beforeEach() {
        // Initialize an empty chessboard
        chessboard = createEmptyChessboard()

        // Place a white Empress on E4
        putPiece(chessboard, positionE4, pieces.whiteEmpress)
    }

    @Test("An Empress can move horizontally")
    testCanMoveHorizontally() {
        // Check the following moves are possible: moveE4_H4, moveE4_A4 
        Expect(isPossible.empressMove(chessboard, moveE4_H4, true)).toBeTruthy()
        Expect(isPossible.empressMove(chessboard, moveE4_A4, true)).toBeTruthy()
    }

    @Test("A Empress can move vertically")
    testCanMoveVertically() {
        // Check the following moves are possible: moveE4_E1, moveE4_E8
        Expect(isPossible.empressMove(chessboard, moveE4_E1, true)).toBeTruthy()
        Expect(isPossible.empressMove(chessboard, moveE4_E8, true)).toBeTruthy()        
    }

    @Test("An Empress can move two squares horizontally and one square vertically")
    testCanMoveTwoHorizontalAndOneVertical() {
        // Check the following moves are possible: 
        // moveE4_G3
        Expect(isPossible.empressMove(chessboard, moveE4_G3, true)).toBeTruthy()

        // moveE4_G5
        Expect(isPossible.empressMove(chessboard, moveE4_G5, true)).toBeTruthy()

        // moveE4_C3
        Expect(isPossible.empressMove(chessboard, moveE4_C3, true)).toBeTruthy()

        // moveE4_C5
        Expect(isPossible.empressMove(chessboard, moveE4_C5, true)).toBeTruthy()
    }

    @Test("An Empress can move two squares vertically  and one square horizontally")
    testCanMoveTwoVerticalAndOneHorizontal() {
        // moveE4_F2
        Expect(isPossible.empressMove(chessboard, moveE4_F2, true)).toBeTruthy()

        // moveE4_F6
        Expect(isPossible.empressMove(chessboard, moveE4_F6, true)).toBeTruthy()

        // moveE4_D2
        Expect(isPossible.empressMove(chessboard, moveE4_D2, true)).toBeTruthy()

        // moveE4_D6
        Expect(isPossible.empressMove(chessboard, moveE4_D6, true)).toBeTruthy()  
    }


    @Test("A Empress cannot move diagonally")
    testCannotMoveDiagonally() {
        // Check the following moves are impossible: 
        // moveE4_A8, moveE4_B1, moveE4_H7, moveE4_H1
        Expect(isPossible.empressMove(chessboard, moveE4_A8, true)).not.toBeTruthy()
        Expect(isPossible.empressMove(chessboard, moveE4_B1, true)).not.toBeTruthy()
        Expect(isPossible.empressMove(chessboard, moveE4_H7, true)).not.toBeTruthy()
        Expect(isPossible.empressMove(chessboard, moveE4_H1, true)).not.toBeTruthy()
    }

    @Test("A Empress can capture a piece from different color")
    testCanCaptureDifferentColor() {
        // Place a black Pawn on H4
        putPiece(chessboard, positionH4, pieces.blackPawn)

        // Check the move moveE4_H4 is possible
        Expect(isPossible.empressMove(chessboard, moveE4_H4, true)).toBeTruthy()        
    }

    @Test("A Empress cannot capture a piece from the same color")
    testCannotCaptureSameColor() {
        // Place a white Pawn on H4
        putPiece(chessboard, positionH4, pieces.whitePawn)

        // Check the move moveE4_H4 is impossible 
        Expect(isPossible.empressMove(chessboard, moveE4_H4, true)).not.toBeTruthy()
    }

    @Test("A Empress cannot leap other pieces, when moving horizontally")
    testCannotLeapHorizontally() {
        // Place a black Pawn on F4
        putPiece(chessboard, positionF4, pieces.blackPawn)

        // Check the move moveE4_H4 is impossible 
        Expect(isPossible.empressMove(chessboard, moveE4_H4, true)).not.toBeTruthy()
    }

    @Test("A Empress cannot leap other pieces, when moving horizontally")
    testCannotLeapvertically() {
        // Place a black Pawn on E3
        putPiece(chessboard, positionE3, pieces.blackPawn)

        // Check the move moveE4_E1 is impossible
        Expect(isPossible.empressMove(chessboard, moveE4_E1, true)).not.toBeTruthy()        
    }
}
